@@ -1,11 +1,11 @@
 from pyrosm.data_manager import get_osm_data
 from pyrosm.frames import prepare_geodataframe
 from pyrosm.utils import validate_custom_filter
-import geopandas as gpd
 import warnings
 
 
-def get_building_data(node_coordinates, way_records, relations, tags_as_columns, custom_filter):
+def get_building_data(node_coordinates, way_records, relations,
+                      tags_as_columns, custom_filter, bounding_box):
     # If custom_filter has not been defined, initialize with default
     if custom_filter is None:
         custom_filter = {"building": True}
@@ -29,12 +29,13 @@ def get_building_data(node_coordinates, way_records, relations, tags_as_columns,
 
     # If there weren't any data, return empty GeoDataFrame
     if nodes is None and ways is None and relations is None:
-        warnings.warn("Could not find any landuse elements for given area.",
+        warnings.warn("Could not find any building elements for given area.",
                       UserWarning,
                       stacklevel=2)
-        return gpd.GeoDataFrame()
+        return None
 
     # Prepare GeoDataFrame
     gdf = prepare_geodataframe(nodes, node_coordinates, ways,
-                               relations, relation_ways, tags_as_columns)
+                               relations, relation_ways, tags_as_columns,
+                               bounding_box)
     return gdf
