@@ -4,9 +4,12 @@ from pyrosm._arrays import concatenate_dicts_of_arrays
 from pyrosm.geometry import create_node_coordinates_lookup
 from pyrosm.frames import create_nodes_gdf
 from pyrosm.utils import validate_custom_filter, validate_osm_keys, \
-    validate_tags_as_columns, validate_booleans
-from shapely.geometry import Polygon, MultiPolygon
-
+    validate_tags_as_columns, validate_booleans, validate_boundary_type, \
+    validate_bounding_box, validate_input_file, get_bounding_box
+from pyrosm.utils.download import get_file_size
+from shapely.geometry import Polygon, MultiPolygon, \
+    LineString, LinearRing, MultiLineString
+from pyrosm.boundary import get_boundary_data
 from pyrosm.buildings import get_building_data
 from pyrosm.landuse import get_landuse_data
 from pyrosm.natural import get_natural_data
@@ -17,6 +20,8 @@ from pyrosm.user_defined import get_user_defined_data
 
 class OSM:
     from pyrosm.utils._compat import PYGEOS_SHAPELY_COMPAT
+    allowed_bbox_types = [Polygon, MultiPolygon, MultiLineString,
+                          LineString, LinearRing]
 
     def __init__(self, filepath,
                  bounding_box=None):
@@ -25,36 +30,41 @@ class OSM:
         ----------
 
         filepath : str
-            Filepath to input OSM dataset ( .osm.pbf | .osm | .osm.bz2 | .osm.gz )
+            Filepath to input OSM dataset ( *.osm.pbf )
 
-        bounding_box : list | shapely.Polygon (optional)
+        bounding_box : list | shapely geometry
             Filtering OSM data spatially is allowed by passing a
             bounding box either as a list `[minx, miny, maxx, maxy]` or
-            as a `shapely.geometry.Polygon`.
+            as a Shapely Polygon/MultiPolygon or closed LineString/LinearRing.
 
-            Note: if using Polygon, the tool will use its bounds.
-            Filtering based on complex shapes is not currently supported.
         """
-        if not isinstance(filepath, str):
-            raise ValueError("'filepath' should be a string.")
-        if not filepath.endswith(".pbf"):
-            raise ValueError(f"Input data should be in Protobuf format (*.osm.pbf). "
-                             f"Found: {filepath.split('.')[-1]}")
 
-        self.filepath = filepath
+        # Check input file
+        self.filepath = validate_input_file(filepath)
 
         if bounding_box is None:
             self.bounding_box = None
-        elif type(bounding_box) in [Polygon, MultiPolygon]:
-            self.bounding_box = bounding_box.bounds
+        elif type(bounding_box) in self.allowed_bbox_types:
+            # Ensures bounding box is a closed geometry
+            # (+ attempts to close MultiLineStrings)
+            self.bounding_box = validate_bounding_box(bounding_box)
         elif isinstance(bounding_box, list):
+            if not len(bounding_box) == 4:
+                raise ValueError("When passing bounding box as a list it should contain 4 coordinates: "
+                                 "[minx, miny, maxx, maxy].")
             self.bounding_box = bounding_box
         else:
-            raise ValueError("bounding_box should be a list or a shapely Polygon.")
+            raise ValueError("bounding_box should be a list, Shapely Polygon or a Shapely LinearRing.")
 
         self.conf = Conf
         self.keep_node_info = False
 
+        # Update file size
+        self.file_size = get_file_size(self.filepath)
+
+        # Get bounding box
+        self._data_bounding_box = get_bounding_box(self.filepath)
+
         # TODO: Add logging as a parameter?
         self._verbose = False
 
@@ -66,11 +76,20 @@ class OSM:
         self._relations = None
 
     def _read_pbf(self):
+        # PBF reading requires a list of bounding box coordinates
+        if type(self.bounding_box) in self.allowed_bbox_types:
+            bounding_box = self.bounding_box.bounds
+        else:
+            bounding_box = self.bounding_box
+
         nodes, ways, relations, way_tags = parse_osm_data(self.filepath,
-                                                          self.bounding_box,
-                                                          exclude_relations=False)
-        self._nodes, self._way_records, \
-        self._relations, self._all_way_tags = nodes, ways, relations, way_tags
+                                             bounding_box,
+                                             exclude_relations=False)
+
+        self._nodes = nodes
+        self._way_records = ways
+        self._relations = relations
+        self._all_way_tags = way_tags
 
         # Prepare node coordinates lookup table
         self._node_coordinates = create_node_coordinates_lookup(self._nodes)
@@ -101,7 +120,7 @@ class OSM:
         elif net_type == "all":
             return None
 
-    def get_network(self, network_type="walking"):
+    def get_network(self, network_type="walking", extra_attributes=None):
         """
         Parses street networks from OSM
         for walking, driving, and cycling.
@@ -110,13 +129,32 @@ class OSM:
         ----------
 
         network_type : str
-            What kind of network to parse. Possible values are: 'walking' | 'cycling' | 'driving' | 'all'.
+            What kind of network to parse.
+            Possible values are:
+              - `'walking'`
+              - `'cycling'`
+              - `'driving'`
+              - `'driving+service'`
+              - `'all'`.
+
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
+
+        See Also
+        --------
+
+        Take a look at the OSM documentation for further details about the data:
+        `https://wiki.openstreetmap.org/wiki/Key:highway <https://wiki.openstreetmap.org/wiki/Key:highway>`__
 
         """
         # Get filter
         network_filter = self._get_network_filter(network_type)
         tags_as_columns = self.conf.tags.highway
 
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
         if self._nodes is None or self._way_records is None:
             self._read_pbf()
 
@@ -124,18 +162,19 @@ class OSM:
         gdf = get_network_data(self._node_coordinates,
                                self._way_records,
                                tags_as_columns,
-                               network_filter
+                               network_filter,
+                               self.bounding_box
                                )
 
         # Do not keep node information unless specifically asked for
         # (they are in a list, and can cause issues when saving the files)
-        if not self.keep_node_info:
+        if not self.keep_node_info and gdf is not None:
             if "nodes" in gdf.columns:
                 gdf = gdf.drop("nodes", axis=1)
 
         return gdf
 
-    def get_buildings(self, custom_filter=None):
+    def get_buildings(self, custom_filter=None, extra_attributes=None):
         """
         Parses buildings from OSM.
 
@@ -143,23 +182,31 @@ class OSM:
         ----------
 
         custom_filter : dict
-            What kind of buildings to parse, see details below.
+            What kind of buildings to parse,
+            see details below.
 
             You can opt-in specific elements by using 'custom_filter'.
             To keep only specific buildings such as 'residential' and 'retail', you can apply
             a custom filter which is a Python dictionary with following format:
-              `custom_filter={'building': ['residential', 'retail']}`
+              - `custom_filter={'building': ['residential', 'retail']}`
 
-        Further info
-        ------------
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
 
-        See OSM documentation for details about the data:
-        https://wiki.openstreetmap.org/wiki/Key:building
+        See Also
+        --------
+
+        Take a look at the OSM documentation for further details about the data:
+        `https://wiki.openstreetmap.org/wiki/Key:building <https://wiki.openstreetmap.org/wiki/Key:building>`__
 
         """
         # Default tags to keep as columns
         tags_as_columns = self.conf.tags.building
 
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
         if self._nodes is None or self._way_records is None:
             self._read_pbf()
 
@@ -167,16 +214,18 @@ class OSM:
                                 self._way_records,
                                 self._relations,
                                 tags_as_columns,
-                                custom_filter)
+                                custom_filter,
+                                self.bounding_box
+                                )
 
         # Do not keep node information unless specifically asked for
         # (they are in a list, and can cause issues when saving the files)
-        if not self.keep_node_info:
+        if not self.keep_node_info and gdf is not None:
             if "nodes" in gdf.columns:
                 gdf = gdf.drop("nodes", axis=1)
         return gdf
 
-    def get_landuse(self, custom_filter=None):
+    def get_landuse(self, custom_filter=None, extra_attributes=None):
         """
         Parses landuse from OSM.
 
@@ -184,18 +233,24 @@ class OSM:
         ----------
 
         custom_filter : dict
-            What kind of landuse to parse, see details below.
+            What kind of landuse to parse,
+            see details below.
 
             You can opt-in specific elements by using 'custom_filter'.
             To keep only specific landuse such as 'construction' and 'industrial', you can apply
             a custom filter which is a Python dictionary with following format:
               `custom_filter={'landuse': ['construction', 'industrial']}`
 
-        Further info
-        ------------
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
+
+        See Also
+        --------
+
+        Take a look at OSM documentation for further details about the data:
+
+        `https://wiki.openstreetmap.org/wiki/Key:landuse <https://wiki.openstreetmap.org/wiki/Key:landuse>`__
 
-        See OSM documentation for details about the data:
-        https://wiki.openstreetmap.org/wiki/Key:landuse
         """
 
         if self._nodes is None or self._way_records is None:
@@ -204,6 +259,10 @@ class OSM:
         # Default tags to keep as columns
         tags_as_columns = self.conf.tags.landuse
 
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
         # If nodes are still in chunks, merge before passing forward
         if isinstance(self._nodes, list):
             self._nodes = concatenate_dicts_of_arrays(self._nodes)
@@ -213,16 +272,18 @@ class OSM:
                                self._way_records,
                                self._relations,
                                tags_as_columns,
-                               custom_filter)
+                               custom_filter,
+                               self.bounding_box
+                               )
 
         # Do not keep node information unless specifically asked for
         # (they are in a list, and can cause issues when saving the files)
-        if not self.keep_node_info:
+        if not self.keep_node_info and gdf is not None:
             if "nodes" in gdf.columns:
                 gdf = gdf.drop("nodes", axis=1)
         return gdf
 
-    def get_natural(self, custom_filter=None):
+    def get_natural(self, custom_filter=None, extra_attributes=None):
         """
         Parses natural from OSM.
 
@@ -230,18 +291,24 @@ class OSM:
         ----------
 
         custom_filter : dict
-            What kind of natural to parse, see details below.
+            What kind of natural to parse,
+            see details below.
 
             You can opt-in specific elements by using 'custom_filter'.
             To keep only specific natural such as 'wood' and 'tree', you can apply
             a custom filter which is a Python dictionary with following format:
               `custom_filter={'natural': ['wood', 'tree']}`
 
-        Further info
-        ------------
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
+
+        See Also
+        --------
+
+        Take a look at OSM documentation for further details about the data:
+
+        `https://wiki.openstreetmap.org/wiki/Key:natural <https://wiki.openstreetmap.org/wiki/Key:natural>`__
 
-        See OSM documentation for details about the data:
-        https://wiki.openstreetmap.org/wiki/Key:natural
         """
 
         if self._nodes is None or self._way_records is None:
@@ -250,6 +317,10 @@ class OSM:
         # Default tags to keep as columns
         tags_as_columns = self.conf.tags.natural
 
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
         # If nodes are still in chunks, merge before passing forward
         if isinstance(self._nodes, list):
             self._nodes = concatenate_dicts_of_arrays(self._nodes)
@@ -259,16 +330,96 @@ class OSM:
                                self._way_records,
                                self._relations,
                                tags_as_columns,
-                               custom_filter)
+                               custom_filter,
+                               self.bounding_box)
 
         # Do not keep node information unless specifically asked for
         # (they are in a list, and can cause issues when saving the files)
-        if not self.keep_node_info:
+        if not self.keep_node_info and gdf is not None:
             if "nodes" in gdf.columns:
                 gdf = gdf.drop("nodes", axis=1)
         return gdf
 
-    def get_pois(self, custom_filter=None):
+    def get_boundaries(self, boundary_type="administrative", name=None, custom_filter=None, extra_attributes=None):
+        """
+        Parses boundaries from OSM.
+
+        Parameters
+        ----------
+
+        boundary_type : str
+            The type of boundaries to parse. Possible values:
+              - `"administrative"` (default)
+              - `"national_park"`
+              - `"political"`
+              - `"postal_code"`
+              - `"protected_area"`
+              - `"aboriginal_lands"`
+              - `"maritime"`
+              - `"lot"`
+              - `"parcel"`
+              - `"tract"`
+              - `"marker"`
+              - `"all"`
+
+        name : str (optional)
+            Name of the administrative area that will be searched for.
+
+        custom_filter : dict (optional)
+            Additional filter for what kind of boundary to parse.
+
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
+
+
+        See Also
+        --------
+
+        Take a look at OSM documentation for further details about the data:
+
+        `https://wiki.openstreetmap.org/wiki/Key:boundary <https://wiki.openstreetmap.org/wiki/Key:bondary>`__
+
+        """
+
+        if self._nodes is None or self._way_records is None:
+            self._read_pbf()
+
+        # Default tags to keep as columns
+        tags_as_columns = self.conf.tags.boundary
+
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
+        # If nodes are still in chunks, merge before passing forward
+        if isinstance(self._nodes, list):
+            self._nodes = concatenate_dicts_of_arrays(self._nodes)
+
+        # Check boundary type
+        boundary_type = validate_boundary_type(boundary_type)
+
+        if name is not None:
+            if not isinstance(name, str):
+                raise ValueError(f"'name' should be text."
+                                 f"Got '{name}' of type {type(name)}.")
+
+        gdf = get_boundary_data(self._node_coordinates,
+                                self._way_records,
+                                self._relations,
+                                tags_as_columns,
+                                custom_filter,
+                                boundary_type,
+                                name,
+                                self.bounding_box)
+
+        # Do not keep node information unless specifically asked for
+        # (they are in a list, and can cause issues when saving the files)
+        if not self.keep_node_info and gdf is not None:
+            if "nodes" in gdf.columns:
+                gdf = gdf.drop("nodes", axis=1)
+        return gdf
+
+    def get_pois(self, custom_filter=None, extra_attributes=None):
         """
         Parse Point of Interest (POI) from OSM.
 
@@ -276,12 +427,15 @@ class OSM:
         ----------
 
         custom_filter : dict
-            An optional custom filter to filter only specific POIs from OpenStreetMap
-            (see details below).
+            An optional custom filter to filter only specific POIs from OpenStreetMap,
+            see details below.
+
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
 
 
-        Filtering by tags
-        -----------------
+        Notes
+        -----
 
         By default, Pyrosm will parse all OSM elements (points, lines and polygons)
         that are associated with following keys:
@@ -299,22 +453,13 @@ class OSM:
         such as supermarkets and book stores by specifying:
           `custom_filter={"amenity": True, "shop": ["supermarket", "books"]}`
 
-        Filtering by OSM Element type
-        -----------------------------
 
-        A specific column called `osm_type` is added to the the resulting GeoDataFrame that informs
-        about the OSM Element type. Possible values are: 'node', 'way' and 'relation'.
-        These values can be used to filter out specific type of elements
-        from the results. If you for example want to keep only POIs that are Points, you can select
-        them with a simple Pandas query:
-        >>> my_poi_gdf = my_poi_gdf.loc[my_poi_gdf['osm_type']=='node']
-
-        Further info
-        ------------
+        See Also
+        --------
 
         You can check the most typical OSM tags for different map features from OSM Wiki
-        https://wiki.openstreetmap.org/wiki/Map_Features . It is also possible to get a quick
-        look at the most typical OSM tags from Pyrosm configuration:
+        `https://wiki.openstreetmap.org/wiki/Map_Features <https://wiki.openstreetmap.org/wiki/Map_Features>`__.
+        It is also possible to get a quick look at the most typical OSM tags from Pyrosm configuration:
 
         >>> from pyrosm.config import Conf
         >>> print("All available OSM keys", Conf.tags.available)
@@ -351,6 +496,10 @@ class OSM:
         for k in custom_filter.keys():
             tags_as_columns += getattr(self.conf.tags, k)
 
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
         # If nodes are still in chunks, merge before passing forward
         if isinstance(self._nodes, list):
             self._nodes = concatenate_dicts_of_arrays(self._nodes)
@@ -360,23 +509,25 @@ class OSM:
                            self._way_records,
                            self._relations,
                            tags_as_columns,
-                           custom_filter)
+                           custom_filter,
+                           self.bounding_box)
 
         # Do not keep node information unless specifically asked for
         # (they are in a list, and can cause issues when saving the files)
-        if not self.keep_node_info:
+        if not self.keep_node_info and gdf is not None:
             if "nodes" in gdf.columns:
                 gdf = gdf.drop("nodes", axis=1)
         return gdf
 
-    def get_osm_by_custom_criteria(self,
-                                   custom_filter,
-                                   osm_keys_to_keep=None,
-                                   filter_type="keep",
-                                   tags_as_columns=None,
-                                   keep_nodes=True,
-                                   keep_ways=True,
-                                   keep_relations=True):
+    def get_data_by_custom_criteria(self,
+                                    custom_filter,
+                                    osm_keys_to_keep=None,
+                                    filter_type="keep",
+                                    tags_as_columns=None,
+                                    keep_nodes=True,
+                                    keep_ways=True,
+                                    keep_relations=True,
+                                    extra_attributes=None):
         """
         Parse OSM data based on custom criteria.
 
@@ -393,6 +544,10 @@ class OSM:
             "keep" | "exclude"
             Whether the filters should be used to keep or exclude the data from OSM.
 
+            Note: "exclude" works only with ways and relations. With nodes, only "keep" is supported,
+            because with "exclude" almost all nodes would always be part of the end result as
+            most nodes do not have tag information.
+
         tags_as_columns : list
             Which tags should be kept as columns in the resulting GeoDataFrame.
 
@@ -405,6 +560,9 @@ class OSM:
         keep_relations : bool
             Whether or not the relations should be kept in the resulting GeoDataFrame if they are found.
 
+        extra_attributes : list (optional)
+            Additional OSM tag keys that will be converted into columns in the resulting GeoDataFrame.
+
         """
 
         # Check that the custom filter is in correct format
@@ -440,6 +598,10 @@ class OSM:
             # Validate tags
             validate_tags_as_columns(tags_as_columns)
 
+        if extra_attributes is not None:
+            validate_tags_as_columns(extra_attributes)
+            tags_as_columns += extra_attributes
+
         # Validate booleans
         validate_booleans(keep_nodes, keep_ways, keep_relations)
 
@@ -460,12 +622,13 @@ class OSM:
                                     filter_type,
                                     keep_nodes,
                                     keep_ways,
-                                    keep_relations
+                                    keep_relations,
+                                    self.bounding_box
                                     )
 
         # Do not keep node information unless specifically asked for
         # (they are in a list, and can cause issues when saving the files)
-        if not self.keep_node_info:
+        if not self.keep_node_info and gdf is not None:
             if "nodes" in gdf.columns:
                 gdf = gdf.drop("nodes", axis=1)
         return gdf
